from pathlib import Path

import papermill as pm
import pytest

from .switches import anl_RUN_NOTEBOOKS


def notebooks():
    base = Path(__file__).parent.parent / "notebooks"
    return list(_ for _ in base.glob("**/*.ipynb") if "checkpoints" not in str(_))


@pytest.mark.skipif(
    condition=not anl_RUN_NOTEBOOKS,
    reason="Environment set to ignore running notebook tests. See switches.py",
)
@pytest.mark.parametrize("notebook", notebooks())
def test_notebook(notebook):
    base = Path(__file__).parent.parent / "notebooks"
    dst = notebook.relative_to(base)
    dst = Path(__file__).parent / "tmp_notebooks" / str(dst)
    dst.parent.mkdir(exist_ok=True, parents=True)
    pm.execute_notebook(
        notebook,
        dst,
    )



"""
The implementation file for all entities needed for ANAC-ANL 2024

Participants need to provide a class inherited from `SAONegotiator` that implements its __call__ method

"""
from .runner import *

__all__ = runner.__all__



import random
from pathlib import Path
from typing import Any, Callable, Iterable, Sequence

import numpy as np
from negmas.helpers.misc import intin
from negmas.helpers.strings import unique_name
from negmas.inout import Scenario, UtilityFunction, pareto_frontier
from negmas.negotiators import Negotiator
from negmas.outcomes import make_issue, make_os
from negmas.preferences import LinearAdditiveUtilityFunction as U
from negmas.preferences.generators import generate_utility_values
from negmas.preferences.ops import nash_points
from negmas.preferences.value_fun import TableFun
from negmas.sao.mechanism import SAOMechanism
from negmas.tournaments.neg.simple import SimpleTournamentResults, cartesian_tournament

from anl.anl2024.negotiators.builtins import (
    Boulware,
    Conceder,
    Linear,
    MiCRO,
    NaiveTitForTat,
)

# from anl.anl2024.negotiators.builtin import (
#     StochasticBoulware,
#     StochasticConceder,
#     StochasticLinear,
# )

__all__ = [
    "anl2024_tournament",
    "mixed_scenarios",
    "pie_scenarios",
    "arbitrary_pie_scenarios",
    "monotonic_pie_scenarios",
    "zerosum_pie_scenarios",
    "ScenarioGenerator",
    "DEFAULT_AN2024_COMPETITORS",
    "DEFAULT_TOURNAMENT_PATH",
    "DEFAULT2024SETTINGS",
]

DEFAULT_AN2024_COMPETITORS = (
    Conceder,
    Linear,
    Boulware,
    NaiveTitForTat,
    MiCRO,
    # StochasticLinear,
    # StochasticConceder,
    # StochasticBoulware,
)
"""Default set of negotiators (agents) used as competitors"""

DEFAULT_TOURNAMENT_PATH = Path.home() / "negmas" / "anl2024" / "tournaments"
"""Default location to store tournament logs"""

ReservedRanges = tuple[tuple[float, ...], ...]


DEFAULT2024SETTINGS = dict(
    n_ufuns=2,
    n_scenarios=50,
    n_outcomes=(900, 1100),
    n_steps=(10, 10_000),
    n_repetitions=5,
    reserved_ranges=((0.0, 1.0), (0.0, 1.0)),
    competitors=DEFAULT_AN2024_COMPETITORS,
    rotate_ufuns=True,
    time_limit=60,
    pend=0,
    pend_per_second=0,
    step_time_limit=None,
    negotiator_time_limit=None,
    self_play=True,
    randomize_runs=True,
    known_partner=False,
    final_score=("advantage", "mean"),
    scenario_generator="mix",
    outcomes_log_uniform=True,
    generator_params=dict(
        reserved_ranges=((0.0, 1.0), (0.0, 1.0)),
        log_uniform=False,
        zerosum_fraction=0.05,
        monotonic_fraction=0.25,
        curve_fraction=0.5,
        pareto_first=False,
        n_pareto=(0.005, 0.25),
    ),
)
"""Default settings for ANL 2024"""


ScenarioGenerator = Callable[[int, int | tuple[int, int] | list[int]], list[Scenario]]
"""Type of callable that can be used for generating scenarios. It must receive the number of scenarios and number of outcomes (as int, tuple or list) and return a list of `Scenario` s"""


def pie_scenarios(
    n_scenarios: int = 20,
    n_outcomes: int | tuple[int, int] | list[int] = 100,
    *,
    reserved_ranges: ReservedRanges = ((0.0, 0.999999), (0.0, 0.999999)),
    log_uniform: bool = True,
    monotonic=False,
) -> list[Scenario]:
    """Creates single-issue scenarios with arbitrary/monotonically increasing utility functions

    Args:
        n_scenarios: Number of scenarios to create
        n_outcomes: Number of outcomes per scenario. If a tuple it will be interpreted as a min/max range to sample n. outcomes from.
                    If a list, samples from this list will be used (with replacement).
        reserved_ranges: Ranges of reserved values for first and second negotiators
        log_uniform: If given, the distribution used will be uniform on the logarithm of n. outcomes (only used when n_outcomes is a 2-valued tuple).
        monotonic: If true all ufuns are monotonically increasing in the portion of the pie

    Remarks:
        - When n_outcomes is a tuple, the number of outcomes for each scenario will be sampled independently.
    """
    ufun_sets = []
    base_name = "DivideTyePie" if monotonic else "S"

    def normalize(x):
        mn, mx = x.min(), x.max()
        return ((x - mn) / (mx - mn)).tolist()

    def make_monotonic(x, i):
        x = np.sort(np.asarray(x), axis=None)

        if i:
            x = x[::-1]
        r = random.random()
        if r < 0.33:
            x = np.exp(x)
        elif r < 0.67:
            x = np.log(x)
        else:
            pass
        return normalize(x)

    max_jitter_level = 0.8
    for i in range(n_scenarios):
        n = intin(n_outcomes, log_uniform)
        issues = (
            make_issue(
                [f"{i}_{n-1 - i}" for i in range(n)],
                "portions" if not monotonic else "i1",
            ),
        )
        # funs = [
        #     dict(
        #         zip(
        #             issues[0].all,
        #             # adjust(np.asarray([random.random() for _ in range(n)])),
        #             generate(n, i),
        #         )
        #     )
        #     for i in range(2)
        # ]
        os = make_os(issues, name=f"{base_name}{i}")
        outcomes = list(os.enumerate_or_sample())
        ufuns = U.generate_bilateral(
            outcomes,
            conflict_level=0.5 + 0.5 * random.random(),
            conflict_delta=random.random(),
        )
        jitter_level = random.random() * max_jitter_level
        funs = [
            np.asarray([float(u(_)) for _ in outcomes])
            + np.random.random() * jitter_level
            for u in ufuns
        ]

        if monotonic:
            funs = [make_monotonic(x, i) for i, x in enumerate(funs)]
        else:
            funs = [normalize(x) for x in funs]
        ufuns = tuple(
            U(
                values=(TableFun(dict(zip(issues[0].all, vals))),),
                name=f"{uname}{i}",
                outcome_space=os,
                # reserved_value=(r[0] + random.random() * (r[1] - r[0] - 1e-8)),
            )
            for (uname, vals) in zip(("First", "Second"), funs)
            # for (uname, r, vals) in zip(("First", "Second"), reserved_ranges, funs)
        )
        sample_reserved_values(ufuns, reserved_ranges=reserved_ranges)
        ufun_sets.append(ufuns)

    return [
        Scenario(
            outcome_space=ufuns[0].outcome_space,  # type: ignore We are sure this is not None
            ufuns=ufuns,
        )
        for ufuns in ufun_sets
    ]


def arbitrary_pie_scenarios(
    n_scenarios: int = 20,
    n_outcomes: int | tuple[int, int] | list[int] = 100,
    *,
    reserved_ranges: ReservedRanges = ((0.0, 0.999999), (0.0, 0.999999)),
    log_uniform: bool = True,
) -> list[Scenario]:
    return pie_scenarios(
        n_scenarios,
        n_outcomes,
        reserved_ranges=reserved_ranges,
        log_uniform=log_uniform,
        monotonic=False,
    )


def monotonic_pie_scenarios(
    n_scenarios: int = 20,
    n_outcomes: int | tuple[int, int] | list[int] = 100,
    *,
    reserved_ranges: ReservedRanges = ((0.0, 0.999999), (0.0, 0.999999)),
    log_uniform: bool = True,
) -> list[Scenario]:
    return pie_scenarios(
        n_scenarios,
        n_outcomes,
        reserved_ranges=reserved_ranges,
        log_uniform=log_uniform,
        monotonic=True,
    )


def sample_reserved_values(
    ufuns: tuple[UtilityFunction, ...],
    pareto: tuple[tuple[float, ...], ...] | None = None,
    reserved_ranges: ReservedRanges = ((0.0, 1.0), (0.0, 1.0)),
    eps: float = 1e-3,
) -> tuple[float, ...]:
    """
    Samples reserved values that are guaranteed to allow some rational outcomes for the given ufuns and sets the reserved values.

    Args:
        ufuns: tuple of utility functions to sample reserved values for
        pareto: The pareto frontier. If not given, it will be calculated
        reserved_ranges: the range to sample reserved values from. Notice that the upper limit of this range will be updated
                         to ensure some rational outcoms
        eps: A small number indicating the absolute guaranteed margin of the sampled reserved value from the Nash point.

    """
    n_funs = len(ufuns)
    if pareto is None:
        pareto = pareto_frontier(ufuns)[0]
    assert pareto is not None, f"Cannot find the pareto frontier."
    nash = nash_points(ufuns, frontier=pareto, ranges=[(0, 1) for _ in range(n_funs)])
    if not nash:
        raise ValueError(
            f"Cannot find the Nash point so we cannot find the appropriate reserved ranges"
        )
    nash_utils = nash[0][0]
    if not reserved_ranges:
        reserved_ranges = tuple((0, 1) for _ in range(n_funs))
    reserved_ranges = tuple(
        tuple(min(r[_], n) for _ in range(n_funs))
        for n, r in zip(nash_utils, reserved_ranges)
    )
    reserved = tuple(
        r[0] + (r[1] - eps - r[0]) * random.random() for r in reserved_ranges
    )
    for u, r in zip(ufuns, reserved):
        u.reserved_value = float(r)
    return reserved


def zerosum_pie_scenarios(
    n_scenarios: int = 20,
    n_outcomes: int | tuple[int, int] | list[int] = 100,
    *,
    reserved_ranges: ReservedRanges = ((0.0, 0.499999), (0.0, 0.499999)),
    log_uniform: bool = True,
) -> list[Scenario]:
    """Creates scenarios all of the DivideThePie variety with proportions giving utility

    Args:
        n_scenarios: Number of scenarios to create
        n_outcomes: Number of outcomes per scenario (if a tuple it will be interpreted as a min/max range to sample n. outcomes from).
        reserved_ranges: Ranges of reserved values for first and second negotiators
        log_uniform: If given, the distribution used will be uniform on the logarithm of n. outcomes (only used when n_outcomes is a 2-valued tuple).

    Remarks:
        - When n_outcomes is a tuple, the number of outcomes for each outcome will be sampled independently
    """
    ufun_sets = []
    for i in range(n_scenarios):
        n = intin(n_outcomes, log_uniform)
        issues = (make_issue([f"{i}_{n-1 - i}" for i in range(n)], "portions"),)
        ufuns = tuple(
            U(
                values=(
                    TableFun(
                        {
                            _: float(int(str(_).split("_")[k]) / (n - 1))
                            for _ in issues[0].all
                        }
                    ),
                ),
                name=f"{uname}{i}",
                # reserved_value=(r[0] + random.random() * (r[1] - r[0] - 1e-8)),
                outcome_space=make_os(issues, name=f"DivideTyePie{i}"),
            )
            for k, uname in enumerate(("First", "Second"))
            # for k, (uname, r) in enumerate(zip(("First", "Second"), reserved_ranges))
        )
        sample_reserved_values(
            ufuns,
            pareto=tuple(
                tuple(u(_) for u in ufuns)
                for _ in make_os(issues).enumerate_or_sample()
            ),
            reserved_ranges=reserved_ranges,
        )
        ufun_sets.append(ufuns)

    return [
        Scenario(
            outcome_space=ufuns[0].outcome_space,  # type: ignore We are sure this is not None
            ufuns=ufuns,
        )
        for ufuns in ufun_sets
    ]


def mixed_scenarios(
    n_scenarios: int = DEFAULT2024SETTINGS["n_scenarios"],  # type: ignore
    n_outcomes: int
    | tuple[int, int]
    | list[int] = DEFAULT2024SETTINGS["n_outcomes"],  # type: ignore
    *,
    reserved_ranges: ReservedRanges = DEFAULT2024SETTINGS["reserved_ranges"],  # type: ignore
    log_uniform: bool = DEFAULT2024SETTINGS["outcomes_log_uniform"],  # type: ignore
    zerosum_fraction: float = DEFAULT2024SETTINGS["generator_params"]["zerosum_fraction"],  # type: ignore
    monotonic_fraction: float = DEFAULT2024SETTINGS["generator_params"]["monotonic_fraction"],  # type: ignore
    curve_fraction: float = DEFAULT2024SETTINGS["generator_params"]["curve_fraction"],  # type: ignore
    pareto_first: bool = DEFAULT2024SETTINGS["generator_params"]["pareto_first"],  # type: ignore
    n_ufuns: int = DEFAULT2024SETTINGS["n_ufuns"],  # type: ignore
    n_pareto: int | float | tuple[float | int, float | int] | list[int | float] = DEFAULT2024SETTINGS["generator_params"]["n_pareto"],  # type: ignore
    pareto_log_uniform: bool = False,
    n_trials=10,
) -> list[Scenario]:
    """Generates a mix of zero-sum, monotonic and general scenarios

    Args:
        n_scenarios: Number of scenarios to genearate
        n_outcomes: Number of outcomes (or a list of range thereof).
        reserved_ranges: the range allowed for reserved values for each ufun.
                         Note that the upper limit will be overridden to guarantee
                         the existence of at least one rational outcome
        log_uniform: Use log-uniform instead of uniform sampling if n_outcomes is a tuple giving a range.
        zerosum_fraction: Fraction of zero-sum scenarios. These are original DivideThePie scenarios
        monotonic_fraction: Fraction of scenarios where each ufun is a monotonic function of the received pie.
        curve_fraction: Fraction of general and monotonic scenarios that use a curve for Pareto generation instead of
                        a piecewise linear Pareto frontier.
        pareto_first: If given, the Pareto frontier will always be in the first set of outcomes
        n_ufuns: Number of ufuns to generate per scenario
        n_pareto: Number of outcomes on the Pareto frontier in general scenarios.
                Can be specified as a number, a tuple of a min and max to sample within, a list of possibilities.
                Each value can either be an integer > 1 or a fraction of the number of outcomes in the scenario.
        pareto_log_uniform: Use log-uniform instead of uniform sampling if n_pareto is a tuple
        n_trials: Number of times to retry generating each scenario if failures occures

    Returns:
        A list `Scenario` s
    """
    assert zerosum_fraction + monotonic_fraction <= 1.0
    nongeneral_fraction = zerosum_fraction + monotonic_fraction
    ufun_sets = []
    for i in range(n_scenarios):
        r = random.random()
        n = intin(n_outcomes, log_uniform)
        name = "S"
        if r < nongeneral_fraction:
            n_pareto_selected = n
            name = "DivideThePieGen"
        else:
            if isinstance(n_pareto, Iterable):
                n_pareto = type(n_pareto)(
                    int(_ * n + 0.5) if _ < 1 else int(_) for _ in n_pareto  # type: ignore
                )
            else:
                n_pareto = int(0.5 + n_pareto * n) if n_pareto < 1 else int(n_pareto)
            n_pareto_selected = intin(n_pareto, log_uniform=pareto_log_uniform)  # type: ignore
        for _ in range(n_trials):
            try:
                if r < zerosum_fraction:
                    vals = generate_utility_values(
                        n_pareto_selected,
                        n,
                        n_ufuns=n_ufuns,
                        pareto_first=pareto_first,
                        pareto_generator="zero_sum",
                    )
                    name = "DivideThePie"
                else:
                    if n_pareto_selected < 2:
                        n_pareto_selected = 2
                    vals = generate_utility_values(
                        n_pareto_selected,
                        n,
                        n_ufuns=n_ufuns,
                        pareto_first=pareto_first,
                        pareto_generator="curve"
                        if random.random() < curve_fraction
                        else "piecewise_linear",
                    )
                break
            except:
                continue
        else:
            continue

        issues = (make_issue([f"{i}_{n-1 - i}" for i in range(n)], "portions"),)
        ufuns = tuple(
            U(
                values=(
                    TableFun(
                        {_: float(vals[i][k]) for i, _ in enumerate(issues[0].all)}
                    ),
                ),
                name=f"{uname}{i}",
                # reserved_value=(r[0] + random.random() * (r[1] - r[0] - 1e-8)),
                outcome_space=make_os(issues, name=f"{name}{i}"),
            )
            for k, uname in enumerate(("First", "Second"))
            # for k, (uname, r) in enumerate(zip(("First", "Second"), reserved_ranges))
        )
        sample_reserved_values(ufuns, reserved_ranges=reserved_ranges)
        ufun_sets.append(ufuns)

    return [
        Scenario(
            outcome_space=ufuns[0].outcome_space,  # type: ignore We are sure this is not None
            ufuns=ufuns,
        )
        for ufuns in ufun_sets
    ]


GENERAROR_MAP = dict(
    monotonic=monotonic_pie_scenarios,
    arbitrary=arbitrary_pie_scenarios,
    zerosum=zerosum_pie_scenarios,
    default=mixed_scenarios,
    mix=mixed_scenarios,
)

DEFAULT2024GENERATOR = mixed_scenarios
"""Default generator type for ANL 2024"""


def anl2024_tournament(
    n_scenarios: int = DEFAULT2024SETTINGS["n_scenarios"],  # type: ignore
    n_outcomes: int | tuple[int, int] | list[int] = DEFAULT2024SETTINGS["n_outcomes"],  # type: ignore
    competitors: tuple[type[Negotiator] | str, ...]
    | list[type[Negotiator] | str] = DEFAULT_AN2024_COMPETITORS,
    rotate_ufuns: bool = DEFAULT2024SETTINGS["rotate_ufuns"],  # type: ignore
    n_repetitions: int = DEFAULT2024SETTINGS["n_repetitions"],  # type: ignore
    n_steps: int | tuple[int, int] | None = DEFAULT2024SETTINGS["n_steps"],  # type: ignore
    time_limit: float | tuple[float, float] | None = DEFAULT2024SETTINGS["time_limit"],  # type: ignore
    pend: float | tuple[float, float] = DEFAULT2024SETTINGS["pend"],  # type: ignore
    pend_per_second: float
    | tuple[float, float] = DEFAULT2024SETTINGS["pend_per_second"],  # type: ignore
    step_time_limit: float
    | tuple[float, float]
    | None = DEFAULT2024SETTINGS["step_time_limit"],  # type: ignore
    negotiator_time_limit: float
    | tuple[float, float]
    | None = DEFAULT2024SETTINGS["negotiator_time_limit"],  # type: ignore
    self_play: bool = DEFAULT2024SETTINGS["self_play"],  # type: ignore
    randomize_runs: bool = DEFAULT2024SETTINGS["randomize_runs"],  # type: ignore
    known_partner: bool = DEFAULT2024SETTINGS["known_partner"],  # type: ignore
    final_score: tuple[str, str] = DEFAULT2024SETTINGS["final_score"],  # type: ignore
    scenario_generator: str
    | ScenarioGenerator = DEFAULT2024SETTINGS["scenario_generator"],  # type: ignore
    generator_params: dict[str, Any] | None = DEFAULT2024SETTINGS["generator_params"],  # type: ignore
    competitor_params: Sequence[dict | None] | None = None,
    name: str | None = None,
    nologs: bool = False,
    njobs: int = 0,
    plot_fraction: float = 0.2,
    verbosity: int = 1,
    save_every: int = 0,
    save_stats: bool = True,
    base_path: Path | None = None,
    plot_params: dict[str, Any] | None = None,
    raise_exceptions: bool = True,
) -> SimpleTournamentResults:
    """Runs an ANL 2024 tournament

    Args:
        n_scenarios: Number of negotiation scenarios
        n_outcomes: Number of outcomes (or a min/max tuple of n. outcomes) for each scenario
        competitors: list of competitor agents
        competitor_params: If given, parameters to construct each competitor
        rotate_ufuns: If given, each scenario will be tried with both orders of the ufuns.
        n_repetitions: Number of times to repeat each negotiation
        n_steps: Number of steps/rounds allowed for the each negotiation (None for no-limit and a 2-valued tuple for sampling from a range)
        time_limit: Number of seconds allowed for the each negotiation (None for no-limit and a 2-valued tuple for sampling from a range)
        pend: Probability of ending the negotiation every step/round (None for no-limit and a 2-valued tuple for sampling from a range)
        pend_per_second: Probability of ending the negotiation every second (None for no-limit and a 2-valued tuple for sampling from a range)
        step_time_limit: Time limit for every negotiation step (None for no-limit and a 2-valued tuple for sampling from a range)
        negotiator_time_limit: Time limit for all actions of every negotiator (None for no-limit and a 2-valued tuple for sampling from a range)
        name: Name of the tournament
        nologs: If given, no logs will be saved
        njobs: Number of parallel jobs to use. -1 for serial and 0 for all cores
        plot_fraction: Fraction of negotiations to plot. Only used if not nologs
        verbosity: Verbosity level. The higher the more verbose
        self_play: Allow negotiators to run against themselves.
        randomize_runs: Randomize the order of negotiations
        save_every: Save logs every this number of negotiations
        save_stats: Save statistics for scenarios
        known_partner: Allow negotiators to know the type of their partner (through their ID)
        final_score: The metric and statistic used to calculate the score. Metrics are: advantage, utility, welfare, partner_welfare and Stats are: median, mean, std, min, max
        base_path: Folder in which to generate the logs folder for this tournament. Default is ~/negmas/anl2024/tournaments
        scenario_generator: An alternative method for generating bilateral negotiation scenarios. Must receive the number of scenarios and number of outcomes.
        generator_params: Parameters passed to the scenario generator
        plot_params: If given, overrides plotting parameters. See `nemgas.sao.SAOMechanism.plot()` for all parameters

    Returns:
        Tournament results as a `SimpleTournamentResults` object.
    """
    if generator_params is None:
        generator_params = dict()
    if isinstance(scenario_generator, str):
        scenario_generator = GENERAROR_MAP[scenario_generator]
    all_outcomes = not scenario_generator == zerosum_pie_scenarios
    if nologs:
        path = None
    elif base_path is not None:
        path = Path(base_path) / (name if name else unique_name("anl"))
    else:
        path = DEFAULT_TOURNAMENT_PATH / (name if name else unique_name("anl"))
    params = dict(
        ylimits=(0, 1),
        mark_offers_view=True,
        mark_pareto_points=all_outcomes,
        mark_all_outcomes=all_outcomes,
        mark_nash_points=True,
        mark_kalai_points=all_outcomes,
        mark_max_welfare_points=all_outcomes,
        show_agreement=True,
        show_pareto_distance=False,
        show_nash_distance=True,
        show_kalai_distance=False,
        show_max_welfare_distance=False,
        show_max_relative_welfare_distance=False,
        show_end_reason=True,
        show_annotations=not all_outcomes,
        show_reserved=True,
        show_total_time=True,
        show_relative_time=True,
        show_n_steps=True,
    )
    if plot_params:
        params = params.update(plot_params)
    return cartesian_tournament(
        competitors=tuple(competitors),
        scenarios=scenario_generator(n_scenarios, n_outcomes, **generator_params),
        competitor_params=competitor_params,
        rotate_ufuns=rotate_ufuns,
        n_repetitions=n_repetitions,
        path=path,
        njobs=njobs,
        mechanism_type=SAOMechanism,
        n_steps=n_steps,
        time_limit=time_limit,
        pend=pend,
        pend_per_second=pend_per_second,
        step_time_limit=step_time_limit,
        negotiator_time_limit=negotiator_time_limit,
        mechanism_params=None,
        plot_fraction=plot_fraction,
        verbosity=verbosity,
        self_play=self_play,
        randomize_runs=randomize_runs,
        save_every=save_every,
        save_stats=save_stats,
        final_score=final_score,
        id_reveals_type=known_partner,
        name_reveals_type=True,
        plot_params=params,
        raise_exceptions=raise_exceptions,
    )


if __name__ == "__main__":
    anl2024_tournament(
        # competitors=(StochasticBoulware, StochasticLinear),
        n_scenarios=5,
        n_repetitions=3,
        n_outcomes=10,
        verbosity=2,
        njobs=-1,
    )



import random

from negmas.outcomes import Outcome
from negmas.preferences import PresortingInverseUtilityFunction
from negmas.sao import ResponseType, SAONegotiator, SAOResponse, SAOState
from negmas.sao.negotiators import (
    AspirationNegotiator,
    BoulwareTBNegotiator,
    ConcederTBNegotiator,
    LinearTBNegotiator,
    NaiveTitForTatNegotiator,
)

__all__ = ["Linear", "Conceder", "Boulware", "MiCRO", "NaiveTitForTat"]


class NaiveTitForTat(NaiveTitForTatNegotiator):
    """
    A simple behavioral strategy that assumes a zero-sum game
    """


class StochasticLinear(AspirationNegotiator):
    """
    Time-based linear negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="linear",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class StochasticConceder(AspirationNegotiator):
    """
    Time-based conceder negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="conceder",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class StochasticBoulware(AspirationNegotiator):
    """
    Time-based boulware negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="boulware",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class Linear(LinearTBNegotiator):
    """
    Time-based linear negotiation strategy
    """

    ...


class Conceder(ConcederTBNegotiator):
    """
    Time-based conceder negotiation strategy
    """


class Boulware(BoulwareTBNegotiator):
    """
    Time-based boulware negotiation strategy
    """


class MiCRO(SAONegotiator):
    """
    A simple implementation of the MiCRO negotiation strategy

    Remarks:
        - This is a simplified implementation of the MiCRO strategy.
        - It is not guaranteed to exactly match the published work.
        - MiCRO was introduced here:
          de Jonge, Dave. "An Analysis of the Linear Bilateral ANAC Domains Using the MiCRO Benchmark Strategy."
          Proceedings of the Thirty-First International Joint Conference on Artificial Intelligence, IJCAI. 2022.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # initialize local variables
        self.next_indx: int = 0
        self.sorter = None
        self._received, self._sent = set(), set()

    def __call__(self, state: SAOState) -> SAOResponse:
        """The main implementation of the MiCRO strategy"""
        assert self.ufun
        # initialize the sorter (This should better be done in on_negotiation_start() to allow for reuse but this is not needed in ANL)
        if self.sorter is None:
            # A sorter, sorts a ufun and can be used to get outcomes using their utiility
            self.sorter = PresortingInverseUtilityFunction(
                self.ufun, rational_only=True, eps=-1, rel_eps=-1
            )
            # Initialize the sorter. This is an O(nlog n) operation where n is the number of outcomes
            self.sorter.init()
        offer = state.current_offer
        # check whether the offer I received is acceptable
        response = ResponseType.REJECT_OFFER
        # check if the offer is acceptable (if one is received)

        # find my next offer (or best so far if I am not conceding)
        will_concede = len(self._sent) <= len(self._received)
        if not will_concede:
            outcome = self.sorter.outcome_at(self.next_indx - 1)
        else:
            # If I cannot concede, I will use my best offer so far (last one I sent)
            outcome = self.sorter.outcome_at(self.next_indx)
            if outcome is None and self.next_indx > 0:
                outcome = self.sorter.outcome_at(self.next_indx - 1)
                will_concede = False

        # If I received something, check for acceptance
        if offer is not None:
            self._received.add(offer)
            # The Acceptance Policy of MiCRO
            # accept if the offer is not worse than my next offer if I am conceding or the best so far if I am not
            if self.ufun.is_not_worse(offer, outcome):
                return SAOResponse(ResponseType.ACCEPT_OFFER, offer)

        # I will repeat a past offer in any of the following conditions:
        # 1. My next offer is worse than disagreement
        # 2. I am not ready to concede (i.e. I already sent more unique offers than the unique offers I received)
        # The only way, outcome will still be None is if I have no rational outcomes at all. Should never happen in ANL 2024
        if not will_concede or outcome is None or self.ufun.is_worse(outcome, None):
            return SAOResponse(response, self.sample_sent())
        # I am willing and can concede. Concede by one outcome
        self.next_indx += 1
        self._sent.add(outcome)
        return SAOResponse(response, outcome)

    def sample_sent(self) -> Outcome | None:
        # Get an outcome from the set I sent so far (or my best if I sent nothing)
        if not len(self._sent):
            assert self.sorter is not None
            return self.sorter.best()
        return random.choice(list(self._sent))



from click.testing import CliRunner

from anl.cli import main


def test_main():
    runner = CliRunner()
    result = runner.invoke(main, [])

    assert len(result.output) >= 0
    assert result.exit_code == 0



#!/usr/bin/env python
"""The ANL universal command line tool"""
import math
import sys
from collections import defaultdict
from functools import partial
from pathlib import Path
from time import perf_counter
from typing import Iterable, List

import click
import click_config_file
import negmas
from negmas.helpers import humanize_time, unique_name
from negmas.helpers.inout import load
from negmas.helpers.types import get_class
from rich import print

import anl
from anl import DEFAULT_AN2024_COMPETITORS
from anl.anl2024.runner import (
    DEFAULT2024SETTINGS,
    DEFAULT_TOURNAMENT_PATH,
    anl2024_tournament,
)

n_completed = 0
n_total = 0


def default_log_path():
    """Default location for all logs"""
    return Path.home() / "negmas" / "anl"


DB_FOLDER = default_log_path().parent / "runsdb"
DB_NAME = "rundb.csv"


def save_run_info(
    name: str, log_path: Path, type_: str = "world", path: Path = DB_FOLDER
):
    try:
        path.mkdir(parents=True, exist_ok=True)
        with open(path / DB_NAME, "a") as f:
            f.write(f"{name},{type_},{str(log_path)}\n")
    except Exception as e:
        print(f"Failed to save run info: {str(e)}")


def default_tournament_path():
    """The default path to store tournament run info"""
    return default_log_path() / "tournaments"


def default_world_path():
    """The default path to store world run info"""
    return default_log_path() / "negotiations"


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(
        f"{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})",
        flush=True,
    )


def shortest_unique_names(strs: List[str], sep="."):
    """
    Finds the shortest unique strings starting from the end of each input
    string based on the separator.

    The final strings will only be unique if the inputs are unique.

    Example:
        given ["a.b.c", "d.e.f", "a.d.c"] it will generate ["b.c", "f", "d.c"]
    """
    lsts = [_.split(sep) for _ in strs]
    names = [_[-1] for _ in lsts]
    if len(names) == len(set(names)):
        return names
    locs = defaultdict(list)
    for i, s in enumerate(names):
        locs[s].append(i)
    mapping = {"": ""}
    for s, l in locs.items():
        if len(s) < 1:
            continue
        if len(l) == 1:
            mapping[strs[l[0]]] = s
            continue
        strs_new = [sep.join(lsts[_][:-1]) for _ in l]
        prefixes = shortest_unique_names(strs_new, sep)
        for loc, prefix in zip(l, prefixes):
            x = sep.join([prefix, s])
            if x.startswith(sep):
                x = x[len(sep) :]
            mapping[strs[loc]] = x
    return [mapping[_] for _ in strs]


def nCr(n, r):
    return math.factorial(n) / math.factorial(r) / math.factorial(n - r)


click.option = partial(click.option, show_default=True)


@click.group()
def main():
    pass


@main.command(help="Runs an ANL 2024 tournament")
@click.option(
    "--parallel/--serial",
    default=True,
    help="Run a parallel/serial tournament on a single machine",
)
@click.option(
    "--name",
    "-n",
    default="random",
    help='The name of the tournament. The special value "random" will result in a random name',
)
@click.option(
    "--outcomes",
    "-o",
    default=DEFAULT2024SETTINGS["n_outcomes"] if not isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Number of outcomes in every scenario. If negative or zero, --min-outcomes and --max-outcomes will beused",
)
@click.option(
    "--min-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][0] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Minimum number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--max-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][1] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Max number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--scenarios",
    "-S",
    default=DEFAULT2024SETTINGS["n_scenarios"],  # type: ignore
    type=int,
    help="Number of scenarios to generate",
)
@click.option(
    "--repetitions",
    "-r",
    default=DEFAULT2024SETTINGS["n_repetitions"],  # type: ignore
    type=int,
    help="Number of repetition for each negotiation",
)
@click.option(
    "--competitors",
    default=";".join(_().type_name for _ in DEFAULT2024SETTINGS["competitors"]),  # type: ignore # type: ignore
    help="A semicolon (;) separated list of agent types to use for the competition. You"
    " can also pass the special value default for the default builtin"
    " agents",
)
@click.option(
    "--metric",
    default=DEFAULT2024SETTINGS["final_score"][0],  # type: ignore # type: ignore
    help="The metric to use for evaluating agents. Can be one of: utility, advantage, partner_welfare, welfare",
)
@click.option(
    "--stat",
    default=DEFAULT2024SETTINGS["final_score"][1],  # type: ignore # type: ignore
    help="The statistic applied to the metric to evaluate agents. Can be one of: mean, median, std, min, max",
)
@click.option(
    "--small/--normal",
    default=False,
    help="If --small, then a small tournament will be run and most parameters will be ignored",
)
@click.option(
    "--known-partner/--unknown-partner",
    default=DEFAULT2024SETTINGS["known_partner"],  # type: ignore
    help="Can the agent know its partner type?",
)
@click.option(
    "--save-logs/--no-logs",
    default=True,
    help="Whether to save logs.",
)
@click.option(
    "--timelimit",
    "-t",
    default=DEFAULT2024SETTINGS["time_limit"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=float,
    help="Number of seconds allowed in every negotiation. Negative numbers mean no-limit",
)
@click.option(
    "--min-timelimit",
    default=DEFAULT2024SETTINGS["time_limit"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Minimum number of seconds in every scenario. Only used of --timelimit is zero or negative",
)
@click.option(
    "--max-timelimit",
    default=DEFAULT2024SETTINGS["time_limit"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Max number of seconds in every scenario. Only used of --timelimit is zero or negative",
)
@click.option(
    "--steps",
    "-s",
    default=DEFAULT2024SETTINGS["n_steps"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Number of negotiation rounds allowed in every negotiation. Negative numbers mean no-limit",
)
@click.option(
    "--min-steps",
    default=DEFAULT2024SETTINGS["n_steps"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Minimum number of steps in every scenario. Only used of --steps is zero or negative",
)
@click.option(
    "--max-steps",
    default=DEFAULT2024SETTINGS["n_steps"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Max number of steps in every scenario. Only used of --steps is zero or negative",
)
@click.option(
    "--pend",
    default=DEFAULT2024SETTINGS["pend"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    type=float,
    help="Probability of ending the negotiation at every round",
)
@click.option(
    "--min-pend",
    default=DEFAULT2024SETTINGS["pend"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    help="Minimum pend in every scenario. Only used of --pend is zero or negative",
)
@click.option(
    "--max-pend",
    default=DEFAULT2024SETTINGS["pend"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    help="Max pend in every scenario. Only used of --pend is zero or negative",
)
@click.option(
    "--self-play/--no-self-play",
    default=DEFAULT2024SETTINGS["self_play"],  # type: ignore
    type=bool,
    help="Allow self-play during the tournament",
)
@click.option(
    "--plot",
    "-p",
    default=0.1 if sys.platform in ("darwin", "linux") else 0,
    type=float,
    help="Fraction of negotiations to plot and save",
)
@click.option(
    "--verbosity",
    "-v",
    default=1,
    type=int,
    help="verbosity level",
)
@click.option(
    "--save_every",
    "-e",
    default=1,
    type=int,
    help="Number of negotiations after which to save stats",
)
@click.option(
    "--generator",
    default=DEFAULT2024SETTINGS["scenario_generator"],  # type: ignore
    type=str,
    help="The method to generate scenarios. Default is mix which generates a mix of scenario types containing zero-sum, monotonic and general scenarios",
)
@click.option(
    "--zerosum",
    "-z",
    default=DEFAULT2024SETTINGS["generator_params"].get("zerosum_fraction", 0.05),  # type: ignore
    type=float,
    help="Fraction of zero-sum scenarios (used when generator=mix)",
)
@click.option(
    "--monotonic",
    "-m",
    default=DEFAULT2024SETTINGS["generator_params"].get("monotonic_fraction", 0.25),  # type: ignore
    type=float,
    help="Fraction of monotonic scenarios (used when generator=mix)",
)
@click.option(
    "--curve",
    "-c",
    default=DEFAULT2024SETTINGS["generator_params"].get("curve_fraction", 0.5),  # type: ignore
    type=float,
    help="Fraction of monotonic and general scenarios generated using a Pareto curve not piecewise linear Pareto (used when generator=mix)",
)
@click.option(
    "--rotate/--no-rotate",
    default=min(DEFAULT2024SETTINGS["rotate_ufuns"], 5),  # type: ignore
    help="Rotate utility functions when creating scenarios for the tournament",
)
@click.option(
    "--randomize/--no-randomize",
    default=DEFAULT2024SETTINGS["randomize_runs"],  # type: ignore
    help="Randomize the order of negotiations or not",
)
@click.option(
    "--raise-exceptions/--ignore-exceptions",
    default=True,
    help="Whether to ignore agent exceptions",
)
@click.option(
    "--settings-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to a yaml/json file containing parameters to pass to the generator",
)
@click.option(
    "--path",
    default="",
    help="A path to be added to PYTHONPATH in which all competitors are stored. You can pass a : separated list of "
    "paths on linux/mac and a ; separated list in windows",
)
@click_config_file.configuration_option()
def tournament2024(
    parallel,
    generator,
    name,
    repetitions,
    raise_exceptions,
    competitors,
    verbosity,
    path,
    metric,
    save_logs,
    stat,
    known_partner,
    save_every,
    randomize,
    self_play,
    plot,
    pend,
    min_pend,
    max_pend,
    timelimit,
    min_timelimit,
    max_timelimit,
    steps,
    min_steps,
    max_steps,
    rotate,
    outcomes,
    min_outcomes,
    max_outcomes,
    scenarios,
    small,
    settings_file,
    zerosum,
    monotonic,
    curve,
):
    generator_params = dict()
    # read settings of the scenario generator from the settings file if available
    if settings_file:
        generator_params = load(settings_file)
    # override settings with the appropriate ones based on the generator
    if generator == "mix":
        generator_params.update(
            zerosum_fraction=zerosum,
            monotonic_fraction=monotonic,
            curve_fraction=curve,
        )
    if small:
        scenarios = 3
        steps = 100
        outcomes = 1000
        timelimit = 30

    def read_range(x, min_x, max_x):
        if x > 0:
            return x
        if min_x < 0 and max_x < 0:
            return min_x
        if min_x == max_x:
            return min_x
        return (min_x, max_x)

    steps = read_range(steps, min_steps, max_steps)
    outcomes = read_range(outcomes, min_outcomes, max_outcomes)
    timelimit = read_range(timelimit, min_timelimit, max_timelimit)
    pend = read_range(pend, min_pend, max_pend)
    if len(path) > 0:
        sys.path.append(path)
    if name == "random":
        name = unique_name(base="", rand_digits=0)

    all_competitors = competitors.split(";")

    def find_type_name(stem: str):
        for pre in (
            "",
            "anl.anl2024.negotiators.",
            "negmas.sao.negotiators.",
            "negmas.genius.gnegotiators.",
        ):
            s = pre + stem
            try:
                get_class(s)
                return s
            except:
                pass
        print(f"[red]ERROR[/red] Unknown Competitor Type: {stem}")
        sys.exit()

    for i, cp in enumerate(all_competitors):
        all_competitors[i] = find_type_name(cp)

    if not all_competitors:
        all_competitors = DEFAULT_AN2024_COMPETITORS
    all_competitors_params = [dict() for _ in range(len(all_competitors))]

    print(f"Tournament will be run between {len(all_competitors)} agents: ")
    print(all_competitors)
    if not isinstance(steps, tuple) and steps <= 0:
        steps = None
    if not isinstance(timelimit, tuple) and timelimit <= 0:
        timelimit = None
    if (
        steps is None
        and timelimit is None
        and not isinstance(pend, tuple)
        and pend <= 0.0
    ):
        print(
            f"[red]ERROR[/red] You specified no way to end the negotiation. You MUST pass either --steps, --timelimit or --pend (or the --min, --max versions of them)"
        )
        sys.exit(1)
    print(f"Will use {scenarios} scenarios of {outcomes} outcomes each.")
    print(f"Negotiations will end if any of the following conditions is satisfied:")
    if steps is not None:
        print(f"\tN. Rounds: {steps}")
    if timelimit is not None:
        print(f"\tN. Seconds: {timelimit}")
    if pend is not None and (isinstance(pend, tuple) or pend > 0):
        print(f"\tProbability of ending each round : {pend}")
    tic = perf_counter()
    results = anl2024_tournament(
        n_scenarios=scenarios,
        n_outcomes=outcomes,
        competitors=all_competitors,
        competitor_params=all_competitors_params,
        rotate_ufuns=rotate,
        n_repetitions=repetitions,
        n_steps=steps,
        time_limit=timelimit,
        pend=pend,
        name=name,
        nologs=not save_logs,
        njobs=0 if parallel else -1,
        plot_fraction=plot,
        verbosity=verbosity,
        self_play=self_play,
        randomize_runs=randomize,
        save_every=save_every,
        known_partner=known_partner,
        final_score=(metric, stat),
        scenario_generator=generator,
        generator_params=generator_params,
        raise_exceptions=raise_exceptions,
    )
    if verbosity <= 0:
        print(results.final_scores)
    print(f"Done in {humanize_time(perf_counter() - tic, show_ms=True)}")
    if save_logs:
        print(f"Detailed logs are stored at: {DEFAULT_TOURNAMENT_PATH / name}")


@main.command(help="Prints ANL version and NegMAS version")
def version():
    print(f"anl: {anl.__version__} (NegMAS: {negmas.__version__})")


if __name__ == "__main__":
    main()



import os

__all__ = [
    "anl_RUNALL_TESTS",
    "anl_FASTRUN",
    "anl_RUN2019",
    "anl_RUN2020",
    "anl_RUN2021_TOURNAMENT",
    "anl_RUN2021_ONESHOT",
    "anl_RUN2021_ONESHOT_SYNC",
    "anl_RUN2021_STD",
    "anl_RUN_TUTORIAL2",
    "anl_RUN_GENIUS",
    "anl_RUN_TOURNAMENTS",
    "anl_RUN_STD_TOURNAMENTS",
    "anl_RUN_COLLUSION_TOURNAMENTS",
    "anl_RUN_SABOTAGE_TOURNAMENTS",
    "anl_RUN_TEMP_FAILING",
]


def is_enabled(val: str) -> bool:
    return os.environ.get(val, "").lower() in ("true", "yes")


def isnot_disabled(val: str) -> bool:
    return os.environ.get(val, "").lower() not in ("false", "no")


anl_RUNALL_TESTS = is_enabled("anl_RUNALL_TESTS")
anl_ON_GITHUB = is_enabled("GITHUB_ACTIONS")
anl_FASTRUN = (is_enabled("anl_FASTRUN") or anl_ON_GITHUB) and not anl_RUNALL_TESTS
anl_RUN_TEMP_FAILING = is_enabled("anl_RUN_TEMP_FAILING")
anl_RUN2021_ONESHOT = isnot_disabled("anl_RUN2021_ONESHOT")
anl_RUN2021_STD = isnot_disabled("anl_RUN2021_STD")
anl_RUN_GENIUS = is_enabled("anl_RUN_GENIUS")
anl_RUN_TOURNAMENTS = is_enabled("anl_RUN_TOURNAMENTS")
anl_RUN_STD_TOURNAMENTS = is_enabled("anl_RUN_STD_TOURNAMENTS")
anl_RUN_COLLUSION_TOURNAMENTS = is_enabled("anl_RUN_COLLUSION_TOURNAMENTS")
anl_RUN_SABOTAGE_TOURNAMENTS = is_enabled("anl_RUN_SABOTAGE_TOURNAMENTS")
anl_RUN2021_TOURNAMENT = isnot_disabled("anl_RUN2021_TOURNAMENT")
anl_RUN2021_ONESHOT_SYNC = isnot_disabled("anl_RUN2021_ONESHOT_SYNC")
anl_RUN2019 = is_enabled("anl_RUN2019")
anl_RUN2020 = is_enabled("anl_RUN2020")
anl_RUN_TUTORIAL2 = isnot_disabled("anl_RUN_TUTORIAL2")
anl_RUN_NOTEBOOKS = isnot_disabled("anl_RUN_NOTEBOOKS")
anl_RUN_SCHEDULER = is_enabled("anl_RUN_SCHEDULER")



#!/usr/bin/env python

import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="anl",
    version="0.1.4",
    description="ANAC Automated Negotiations League Platform",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    long_description_content_type="text/x-rst",  # Optional (see note above),
    author="Yasser Mohammad",
    author_email="yasserfarouk@gmail.com",
    url="https://github.com/autoneg/anl",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        # 'Topic :: Utilities',
    ],
    project_urls={
        "Documentation": "https://anl.readthedocs.io/",
        "Changelog": "https://anl.readthedocs.io/en/latest/changelog.html",
        "Issue Tracker": "https://github.com/autoneg/anl/issues",
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.10",
    install_requires=[
        # 'dataclasses; python_version < "3.7"',
        "click",
        "pytest",
        "hypothesis",
        "prettytable",
        "negmas>=0.10.8",
        "tqdm",
        "joblib",
        "jupyter",
        "gif",
    ],
    extras_require={},
    setup_requires=["pytest-runner"],
    entry_points={
        "console_scripts": [
            "anl = anl.cli:main",
        ]
    },
)
